#include "btreewin.h"
#include "ui_btreewin.h"

#include <QApplication>
#include <QTreeWidgetItem>

#include <string>
using namespace std;

BTreeWin * win = nullptr;

/* ---------------------------------------------------------------------- */

const int Min = 2;
const int Max = 2 * Min;

struct Item
{
    int cnt; // 1 <= cnt <= max
    string key [Max+1]; // 0..cnt-1
    Item* ptr [Max+2]; // 0..cnt
    Item ();
};

Item::Item ()
{
    cnt = 0;
    for (int i = 0; i < Max + 1; i++) key[i] = "";
    for (int i = 0; i < Max + 2; i++) ptr[i] = nullptr;
}

/* ---------------------------------------------------------------------- */

bool search (Item * p, string name)
{
    bool result = false;
    while (p != nullptr && ! result)
    {
        int i = 0;
        while (i < p->cnt && p->key[i] < name) i++;
        if (i < p->cnt && p->key[i] == name)
            result = true;
        else
            p = p->ptr[i];
    }
    return true;
}

/* ---------------------------------------------------------------------- */

void enter0 (Item * p, string name)
{
    int i = 0;
    while (i < p->cnt && p->key[i] < name) i++;
    if (i < p->cnt && p->key[i] == name)
    { /* hodnota je jiz ve stromu */ }
    else
    {
        if (p->ptr[i] != nullptr)
        {
            Item * a = p->ptr[i];
            enter0 (a, name);
            if (a->cnt > Max)
            {
                string t = a->key [Min]; /* delici bod */

                /* presun do b */
                Item * b = new Item;
                for (int k = Min+1; k < a->cnt; k++) b->key[k-Min-1] = a->key[k];
                for (int k = Min+2; k < a->cnt+1; k++) b->ptr[k-Min-2] = a->ptr[k];
                b->cnt = a->cnt-1-Min;

                /* vymazani a */
                for (int k = Min+1; k < a->cnt; k++) a->key[k] = "";
                for (int k = Min+2; k < a->cnt+1; k++) a->ptr[k] = nullptr;
                a->cnt = Min;

                /* pridani do p */
                for (int k = p->cnt-1; k >= i; k--) p->key[k+1] = p->key[k];
                for (int k = p->cnt; k >= i+1; k--) p->ptr[k+1] = p->ptr[k];
                p->key[i] = t;
                p->ptr[i+1] = b;
                p->cnt ++;
            }
        }
        else
        {
            /* pridani do listu */
            for (int k = p->cnt-1; k >= i; k--) p->key[k+1] = p->key[k];
            for (int k = p->cnt; k >= i+1; k--) p->ptr[k+1] = p->ptr[k];
            p->key[i] = name;
            p->ptr[i+1] = nullptr;
            p->cnt ++;
        }
    }
}

void enter (Item * & p, string name)
{
    if (p == nullptr)
        p = new Item;

    enter0 (p, name);


        if (p->cnt > Max)
        {
            Item * a = new Item;
            for (int k = 0; k < Min; k++) a->key[k] = p->key[k];
            for (int k = 0; k <= Min; k++) a->ptr[k] = p->ptr[k];
            a->cnt = Min;

            Item * b = new Item;
            for (int k = Min+1; k < p->cnt; k++) b->key[k-Min-1] = p->key[k];
            for (int k = Min+2; k < p->cnt+1; k++) b->ptr[k-Min-2] = p->ptr[k];
            b->cnt = p->cnt-1-Min;


            string t = p->key [Min];
            for (int k = 0; k < p->cnt; k++) p->key[k] = "";
            for (int k = 0; k < p->cnt+1; k++) p->ptr[k] = nullptr;

            p->cnt = 1;
            p->key [0] = t;
            p->ptr[0] = a;
            p->ptr[1] = b;
        }

}

/* ---------------------------------------------------------------------- */

QIcon folderIcon;
QIcon valueIcon;

void displayBranch (QTreeWidgetItem * target, Item * p)
{
    QTreeWidgetItem * node = new QTreeWidgetItem;
    node->setIcon (0, folderIcon);
    target->addChild (node);

    if (p != nullptr)
    {
        string txt = "";
        for (int i = 0; i < p->cnt ; i++)
        {
            if (i != 0) txt = txt + ", ";
            txt = txt + p->key[i];
        }
        node->setText (0, QString::fromStdString (txt));


        for (int i = 0; i < p->cnt ; i++)
        {
            if (p->ptr[i] != nullptr)
                displayBranch (node, p->ptr[i]);

            QTreeWidgetItem * item = new QTreeWidgetItem;
            item->setText (0, QString::fromStdString (p->key[i]));
            item->setIcon (0, valueIcon);
            node->addChild (item);
        }

        int i = p->cnt;
        if (p->ptr[i] != nullptr)
            displayBranch (node, p->ptr[i]);
    }
}

void display (Item * p)
{
    folderIcon = QIcon::fromTheme ("folder");
    valueIcon = QIcon::fromTheme ("info");
    displayBranch (win->ui->treeWidget->invisibleRootItem(), p);
    win->ui->treeWidget->expandAll ();
}

/* ---------------------------------------------------------------------- */

Item * root = nullptr;

void run ()
{
    enter (root, "10");
    enter (root, "05");
    enter (root, "20");
    enter (root, "07");
    enter (root, "15");

    // for (int i = 20; i <= 80; i++)
    //    enter (root, QString::number(i).toStdString());

    display (root);
}

/* ---------------------------------------------------------------------- */

BTreeWin::BTreeWin(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::BTreeWin)
{
    ui->setupUi(this);
    win = this;
    run ();
}

BTreeWin::~BTreeWin()
{
    delete ui;
}

int main(int argc, char *argv[])
{
    QApplication a(argc, argv);
    BTreeWin w;
    w.show();

    return a.exec();
}

